/* public domain, do as you wish */
#include <X11/Xlib.h>
#include <X11/Xproto.h>
#include <X11/Xatom.h>
#include <X11/extensions/Xrandr.h>

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <dev/wscons/wsconsio.h>

int consfd;

void
setbrightness(int percent)
{
	struct wsdisplay_param param;
	memset(&param, 0, sizeof(param));
#if 0
	param.param = WSDISPLAYIO_PARAM_BACKLIGHT;
#else
	param.param = WSDISPLAYIO_PARAM_BRIGHTNESS;
#endif
	ioctl(consfd, WSDISPLAYIO_GETPARAM, &param);
	param.curval = param.min + param.max * percent / 100;
	ioctl(consfd, WSDISPLAYIO_SETPARAM, &param);
}

double
getbrightness(Display *dpy, int screen, Window root)
{
	double brightness = 0;
	XImage *image;
	int h = DisplayHeight(dpy, screen);
	int w = DisplayWidth(dpy, screen);
	int size;
	
	image = XGetImage(dpy, root, 0, 0, w, h, AllPlanes, ZPixmap);
	size = image->bytes_per_line * image->height;
	for (int i = 0; i < h; i++) {
		unsigned char *line = image->data + i * image->bytes_per_line;
		for (int j = 0; j < w; j++) {
			unsigned char *pix = line + j * image->bits_per_pixel / 8;
			unsigned char r = pix[0];
			unsigned char g = pix[1];
			unsigned char b = pix[2];
			unsigned char a = pix[3];

			brightness += r + b * 1.2 + g * 1.5;
		}
	}
	XDestroyImage(image);

	return brightness / (w * h);
}

void
adjustgamma(Display *dpy, Window root, double cutoff)
{
	XRRScreenResources *res = XRRGetScreenResourcesCurrent(dpy, root);

	int num_crtcs = res->ncrtc;
	for (int c = 0; c < res->ncrtc; c++) {
		int crtcxid = res->crtcs[c];
		XRRCrtcInfo *crtc_info = XRRGetCrtcInfo(dpy, res, crtcxid);
		int size = XRRGetCrtcGammaSize(dpy, crtcxid);
		XRRCrtcGamma *crtc_gamma = XRRAllocGamma(size);

		double scale = 1.0;
		for (int i = 0; i < size; i++) {
			double g = 65535.0 * i / size * cutoff;
			if (i > size * cutoff)
				g += 65535.0 * (1 - cutoff);
			g *= scale;
			crtc_gamma->red[i] = g;
			crtc_gamma->green[i] = g;
			crtc_gamma->blue[i] = g;
		}
		XRRSetCrtcGamma(dpy, crtcxid, crtc_gamma);

		XFree(crtc_gamma);
	}
}

int
main(int argc, char **argv)
{
	Display *dpy = XOpenDisplay(NULL);
	int screen = DefaultScreen(dpy);
	Window root = RootWindow(dpy, screen);
	XEvent ev;
	int dim = 0;
	int olddim = -1;
	double brightness;

	consfd = open("/dev/ttyC0", O_WRONLY);

	XSelectInput(dpy, root, FocusChangeMask);

	while (1) {
		XNextEvent(dpy, &ev);

		brightness = getbrightness(dpy, screen, root);

		if (brightness > 400)
			dim = 1;
		else
			dim = 0;

		if (dim != olddim) {
			setbrightness(dim ? 30 : 60);

			adjustgamma(dpy, root, dim ? 0.5 : 1.0);
		}
		olddim = dim;
	}
}
